﻿local AddonName, ns = ...;
local Astrolabe = DongleStub("Astrolabe-1.0")
local MapMarker = ns:GetModule("MapMarker");
local MapData = LibStub("LibMapData-1.0");

--NOTE: Documentation out of date.
--[[
	How Minimap marking works:
		1. The minimap marker firstly creates 3 tables: minimapMasks, minimapFrames, and MapMarker.activeMinimapRoutes
			* Each entry in minimapFrames is a frame with the minimap route drawn on it and
				dimensions equal to the first two entries in the first numerical index in mapData.
				Each numerical entry contains the dimensions (width/height) of a floor (if no floors, then
				it's the dimensions of the zone's basic map) in the first two indices. I don't know what the
				other indices are for. Look in LibMapData-1.0 for more detail (mapData is based off
				LibMapData-1.0's mapData table).
			* Each entry in minimapMasks is a frame using a "ScrollFrame" template.
				Basically, it's a cheat to get a quick mask for the minimap. For some reason, if only a single
				frame is used to mask all the entries in minimapFrames, it doesn't work and only the most-recently
				created minimapFrame is masked.
			* Each entry in MapMarker.activeMinimapRoutes is a zone[route] that should be displayed on the minimap when in the
				appropriate zone. Every time it's called, MapMarker.updateMinimapRoutes() checks the current zone
				against MapMarker.activeMinimapRoutes's zonelist, and calls :Show() and updates the position of each frame in
				minimapFrames that is also in MapMarker.activeMinimapRoutes

		2. After creating those two tables, whenever a MapMarker.drawMinimapLines(...) is invoked, it creates or
			shows (whichever is more appropriate) the frame for the input route for the input zone.

		3. MinimapFrames are then kept current by MapMarker.updateMinimapRoutes, which repositions each one so that the
			frame is offset by the player's current position, thus reflecting the player's motion through the zone.
]]

--Globals to local
local pairs, next = pairs, next
local tinsert, tremove = tinsert, tremove
local floor = math.floor
local format = string.format
local math_abs = math.abs
local math_sin = math.sin
local math_cos = math.cos
local WorldMapButton = WorldMapButton
--local Minimap = Minimap
local GetPlayerFacing = GetPlayerFacing

local update_dist = 1;
local MINIMAP_LINES = {};
local USED_MINIMAP = {};
local activezoneid;
local activeroute;
local throttle = CreateFrame("Frame", nil, Minimap); --used for OnUpdate

--function declarations
local updateLines

--MINIMAP HELPERS
local MinimapShapes = {
	-- quadrant booleans (same order as SetTexCoord)
	-- {upper-left, lower-left, upper-right, lower-right}
	-- true = rounded, false = squared
	["ROUND"]                 = { true,  true,  true,  true},
	["SQUARE"]                = {false, false, false, false},
	["CORNER-TOPLEFT"]        = { true, false, false, false},
	["CORNER-TOPRIGHT"]       = {false, false,  true, false},
	["CORNER-BOTTOMLEFT"]     = {false,  true, false, false},
	["CORNER-BOTTOMRIGHT"]    = {false, false, false,  true},
	["SIDE-LEFT"]             = { true,  true, false, false},
	["SIDE-RIGHT"]            = {false, false,  true,  true},
	["SIDE-TOP"]              = { true, false,  true, false},
	["SIDE-BOTTOM"]           = {false,  true, false,  true},
	["TRICORNER-TOPLEFT"]     = { true,  true,  true, false},
	["TRICORNER-TOPRIGHT"]    = { true, false,  true,  true},
	["TRICORNER-BOTTOMLEFT"]  = { true,  true, false,  true},
	["TRICORNER-BOTTOMRIGHT"] = {false,  true,  true,  true},
}

local minimap_radius
local minimap_rotate
local indoors = "indoor"

local MinimapSize = { -- radius of minimap
	indoor = {
		[0] = 150,
		[1] = 120,
		[2] = 90,
		[3] = 60,
		[4] = 40,
		[5] = 25,
	},
	outdoor = {
		[0] = 233 + 1/3,
		[1] = 200,
		[2] = 166 + 2/3,
		[3] = 133 + 1/3,
		[4] = 100,
		[5] = 66 + 2/3,
	},
}

local function is_round( dx, dy )
	local map_shape = GetMinimapShape and GetMinimapShape() or "ROUND"

	local q = 1
	if dx > 0 then q = q + 2 end -- right side
	-- XXX Tripple check this
	if dy > 0 then q = q + 1 end -- bottom side

	return MinimapShapes[map_shape][q]
end

local function is_inside( sx, sy, cx, cy, radius )
	local dx = sx - cx
	local dy = sy - cy

	if is_round( dx, dy ) then
		return dx*dx+dy*dy <= radius*radius
	else
		return math_abs( dx ) <= radius and math_abs( dy ) <= radius
	end
end

local function GetFacing()
	if GetPlayerFacing then return GetPlayerFacing() end
	return -MiniMapCompassRing:GetFacing()
end

local last_X, last_Y, last_facing --= 1/0, 1/0, 1/0
local lastzoom;

-- implementation of cache - use zone in the key for an unique identifier
-- because every zone has a different X/Y location and possible yardsizes
local cache_zone, cache_zoneW, cache_zoneH
local X_cache = {}
local Y_cache = {}
local XY_cache_mt = {
	__index = function(t, key)
		local zone, coord = (';'):split( key )
		if cache_zone ~= zone then
			cache_zoneW, cache_zoneH = MapData:MapArea(tonumber(zone))
			cache_zone = zone
		end
		local X = cache_zoneW * floor(coord / 10000) / 10000
		local Y = cache_zoneH * (coord % 10000) / 10000
		X_cache[key] = X
		Y_cache[key] = Y

		-- figure out which one to return
		if t == X_cache then return X else return Y end
	end
}

setmetatable( X_cache, XY_cache_mt )
setmetatable( Y_cache, XY_cache_mt )
--END MINIMAP HELPERS

function MapMarker.CVAR_UPDATE(event, cvar, value)
	if cvar == "ROTATE_MINIMAP" then minimap_rotate = (value == "1") end
end

function MapMarker.MINIMAP_UPDATE_ZOOM()
	local zoom = Minimap:GetZoom()
	if GetCVar("minimapZoom") == GetCVar("minimapInsideZoom") then
		Minimap:SetZoom(zoom < 2 and zoom + 1 or zoom - 1)
	end
	indoors = GetCVar("minimapZoom")+0 == Minimap:GetZoom() and "outdoor" or "indoor"
	Minimap:SetZoom(zoom)
end

function MapMarker.MINIMAP_ZONE_CHANGED()
	drawLines(true)
end

function MapMarker.VARIABLES_LOADED(name)
	minimap_rotate = GetCVar("rotateMinimap") == "1"
end

MapMarker:RegisterEvent("CVAR_UPDATE")
MapMarker:RegisterEvent("MINIMAP_UPDATE_ZOOM")
MapMarker:RegisterEvent("MINIMAP_ZONE_CHANGED")
MapMarker:RegisterEvent("VARIABLES_LOADED")

-- Hides the minimap frame for the input zone with the input routeName
function MapMarker.hideMinimapRoute()
	--MapMarker.undrawLine(Minimap, MINIMAP_LINES, USED_MINIMAP);
	activezoneid = nil;
	activeroute = nil;
	updateLines(true);
	throttle:Hide();
end

-- draws the input route on the minimap
-- 	not really, what it actually does is it draws the input route on a frame and then overlays that atop the minimap with a mask so only the part that is actually atop the minimap is visible to the player. The other way is easier to say though
function MapMarker.drawMinimapLines(zoneid, route)
	activezoneid = zoneid;
	activeroute = route;
	updateLines(true)
	throttle:Show();
end 

updateLines = function(forceRedraw)
	if not activeroute and activezoneid then
		MapMarker.undrawLine(Minimap, MINIMAP_LINES, USED_MINIMAP);
		return
	end

	local _x, _y = GetPlayerMapPosition("player")

	-- invalid coordinates - clear map
	if not _x or not _y or _x < 0 or _x > 1 or _y < 0 or _y > 1 then
		MapMarker.undrawLine(Minimap, MINIMAP_LINES, USED_MINIMAP);
		return
	end

	local zoneW, zoneH = MapData:MapArea(activezoneid)
	if not zoneW or zoneW == 0 then 
		MapMarker.undrawLine(Minimap, MINIMAP_LINES, USED_MINIMAP);
		return 
	end
	local cx, cy = zoneW * _x, zoneH * _y

	local facing, sin, cos
	if minimap_rotate then
		facing = GetFacing()
	end
	
	--Force updates when zoom changes, hooking this was being stubborn
	if(Minimap:GetZoom() ~= lastzoom) then
		forceUpdate = true
		lastzoom = Minimap:GetZoom()
	end
	
	if (not forceUpdate) and facing == last_facing and (last_X-cx)^2 + (last_Y-cy)^2 < update_dist^2 then
		-- no update!
		return
	end
	
	--TODO: Make sure this is sufficent.
	if activezoneid ~= GetCurrentMapAreaID() then
		-- we are viewing a map that isn't the current zone (usually a continent
		-- map), so the coordinates are wrong, unless we translate them
		MapMarker.undrawLine(Minimap, MINIMAP_LINES, USED_MINIMAP);
		return
	end

	MapMarker.undrawLine(Minimap, MINIMAP_LINES, USED_MINIMAP);

	last_X = cx
	last_Y = cy
	last_facing = facing

	if minimap_rotate then
		sin = math_sin(facing)
		cos = math_cos(facing)
	end

	minimap_radius = MinimapSize[indoors][Minimap:GetZoom()]
	local radius = minimap_radius
	local radius2 = radius * radius

	local minX = cx - radius
	local maxX = cx + radius
	local minY = cy - radius
	local maxY = cy + radius

	local div_by_zero_nudge = 0.000001

	local minimap_w = Minimap:GetWidth()
	local minimap_h = Minimap:GetHeight()
	local scale_x = minimap_w / (radius*2)
	local scale_y = minimap_h / (radius*2)

	local minimapScale = Minimap:GetScale()
	
	if type(activeroute) == "table" and #activeroute > 1 then
		-- store color/width
		--local width = (route_data.width_minimap or defaults.width_minimap) / (minimapScale)
		local color = MapMarker.getColor({"Herb"}); --TODO: replace with meaningful color

		-- unless we show hidden
		if true then --TODO: if not hidden
			-- some state data
			local last_x = nil
			local last_y = nil
			local last_inside = nil
			
			-- Set end node as previous to loop route
			local key = format("%s;%s", activezoneid, activeroute[ #activeroute ])
			last_x, last_y = X_cache[key], Y_cache[key]
			if minimap_rotate then
				local dx = last_x - cx
				local dy = last_y - cy
				last_x = cx + dx*cos - dy*sin
				last_y = cy + dx*sin + dy*cos
			end
			last_inside = is_inside( last_x, last_y, cx, cy, radius )

			-- loop over the route
			for i = 1, #activeroute do
				local point = activeroute[i]
				local cur_x, cur_y, cur_inside

				-- if we have a 'fake point' (gap) - clear current values
				--[[ WTF is this?
				
				if point == defaults.fake_point then
					cur_x = nil
					cur_y = nil
					cur_inside = false
				else
				]]
					local key = format("%s;%s", activezoneid, point)
					cur_x, cur_y = X_cache[key], Y_cache[key]
					if minimap_rotate then
						local dx = cur_x - cx
						local dy = cur_y - cy
						cur_x = cx + dx*cos - dy*sin
						cur_y = cy + dx*sin + dy*cos
					end
					cur_inside = is_inside( cur_x, cur_y, cx, cy, radius )
				--end

				-- check if we have any nil values (we cant draw) and check boundingbox
				if cur_x and cur_y and last_x and last_y and not (
					( cur_x < minX and last_x < minX ) or
					( cur_x > maxX and last_x > maxX ) or
					( cur_y < minY and last_y < minY ) or
					( cur_y > maxY and last_y > maxY )
				)
				then
					-- default all to not drawing
					local draw_sx = nil
					local draw_sy = nil
					local draw_ex = nil
					local draw_ey = nil

					-- both inside - easy! draw
					if cur_inside and last_inside then
						draw_sx = last_x
						draw_sy = last_y
						draw_ex = cur_x
						draw_ey = cur_y
					else
						-- direction of line
						local dx = last_x - cur_x
						local dy = last_y - cur_y

						-- calculate point on perpendicular line
						local zx = cx - dy
						local zy = cy + dx

						-- nudge it a bit so we dont get div by 0 problems
						if dx == 0 then dx = div_by_zero_nudge end
						if dy == 0 then dy = div_by_zero_nudge end

						-- calculate intersection point
						local nd = ((cx   -last_x)*(cy-zy) - (cx-zx)*(cy   -last_y)) /
								   ((cur_x-last_x)*(cy-zy) - (cx-zx)*(cur_y-last_y))

						-- perpendicular point (closest to center on the line given)
						local px = last_x + nd * -dx
						local py = last_y + nd * -dy

						-- check range of intersect point
						local dpc_x = cx - px
						local dpc_y = cy - py

						-- distance^2 of the perpendicular point
						local lenpc = dpc_x*dpc_x + dpc_y*dpc_y

						-- the line can only intersect if the perpendicular point is at
						-- least closer than the furthest away point (one of the corners)
						if lenpc < 2*radius2 then

							-- if inside - ready to draw
							if cur_inside then
								draw_ex = cur_x
								draw_ey = cur_y
							else
								-- if we're not inside we can still be in the square - if so dont do any intersection
								-- calculations yet
								if math_abs( cur_x - cx ) < radius and math_abs( cur_y - cy ) < radius then
									draw_ex = cur_x
									draw_ey = cur_y
								else
									-- need to intersect against the square
									-- likely x/y to intersect with
									local minimap_cur_x  = cx + radius * (dx < 0 and 1 or -1)
									local minimap_cur_y  = cy + radius * (dy < 0 and 1 or -1)

									-- which intersection is furthest?
									local delta_cur_x = (minimap_cur_x - cur_x) / -dx
									local delta_cur_y = (minimap_cur_y - cur_y) / -dy

									-- dark magic - needs to be changed to positive signs whenever i can care about it
									if delta_cur_x < delta_cur_y and delta_cur_x < 0 then
										draw_ex = minimap_cur_x
										draw_ey = cur_y + -dy*delta_cur_x
									else
										draw_ex = cur_x + -dx*delta_cur_y
										draw_ey = minimap_cur_y
									end

									-- check if we didn't calculate some wonky offset - has to be inside with
									-- some slack on accuracy
									if math_abs( draw_ex - cx ) > radius*1.01 or
									   math_abs( draw_ey - cy ) > radius*1.01
									then
										draw_ex = nil
										draw_ey = nil
									end
								end

								-- we might have a round corner here - lets see if the quarter with the intersection is round
								if draw_ex and draw_ey and is_round( draw_ex - cx, draw_ey - cy ) then
									-- if we are also within the circle-range
									if lenpc < radius2 then
										-- circle intersection
										local dcx = cx - cur_x
										local dcy = cy - cur_y
										local len_dc = dcx*dcx + dcy*dcy

										local len_d = dx*dx + dy*dy
										local len_ddc = dx*dcx + dy*dcy

										-- discriminant
										local d_sqrt = ( len_ddc*len_ddc - len_d * (len_dc - radius2) )^0.5

										-- calculate point
										draw_ex = cur_x - dx * (-len_ddc + d_sqrt) / len_d
										draw_ey = cur_y - dy * (-len_ddc + d_sqrt) / len_d

										-- have to be on the *same* side of the perpendicular point else it's fake
										if (draw_ex - px)/math_abs(draw_ex - px) ~= (cur_x- px)/math_abs(cur_x - px) or
										   (draw_ey - py)/math_abs(draw_ey - py) ~= (cur_y- py)/math_abs(cur_y - py)
										then
											draw_ex = nil
											draw_ey = nil
										end
									else
										draw_ex = nil
										draw_ey = nil
									end
								end
							end

							-- if inside - ready to draw
							if last_inside then
								draw_sx = last_x
								draw_sy = last_y
							else
								-- if we're not inside we can still be in the square - if so dont do any intersection
								-- calculations yet
								if math_abs( last_x - cx ) < radius and math_abs( last_y - cy ) < radius then
									draw_sx = last_x
									draw_sy = last_y
								else
									-- need to intersect against the square
									-- likely x/y to intersect with
									local minimap_last_x = cx + radius * (dx > 0 and 1 or -1)
									local minimap_last_y = cy + radius * (dy > 0 and 1 or -1)

									-- which intersection is furthest?
									local delta_last_x = (minimap_last_x - last_x) / dx
									local delta_last_y = (minimap_last_y - last_y) / dy

									-- dark magic - needs to be changed to positive signs whenever i can care about it
									if delta_last_x < delta_last_y and delta_last_x < 0 then
										draw_sx = minimap_last_x
										draw_sy = last_y + dy*delta_last_x
									else
										draw_sx = last_x + dx*delta_last_y
										draw_sy = minimap_last_y
									end

									-- check if we didn't calculate some wonky offset - has to be inside with
									-- some slack on accuracy
									if math_abs( draw_sx - cx ) > radius*1.01 or
									   math_abs( draw_sy - cy ) > radius*1.01
									then
										draw_sx = nil
										draw_sy = nil
									end
								end

								-- we might have a round corner here - lets see if the quarter with the intersection is round
								if draw_sx and draw_sy and is_round( draw_sx - cx, draw_sy - cy ) then
									-- if we are also within the circle-range
									if lenpc < radius2 then
										-- circle intersection
										local dcx = cx - cur_x
										local dcy = cy - cur_y
										local len_dc = dcx*dcx + dcy*dcy

										local len_d = dx*dx + dy*dy
										local len_ddc = dx*dcx + dy*dcy

										-- discriminant
										local d_sqrt = ( len_ddc*len_ddc - len_d * (len_dc - radius2) )^0.5

										-- calculate point
										draw_sx = cur_x - dx * (-len_ddc - d_sqrt) / len_d
										draw_sy = cur_y - dy * (-len_ddc - d_sqrt) / len_d

										-- have to be on the *same* side of the perpendicular point else it's fake
										if (draw_sx - px)/math_abs(draw_sx - px) ~= (last_x- px)/math_abs(last_x - px) or
										   (draw_sy - py)/math_abs(draw_sy - py) ~= (last_y- py)/math_abs(last_y - py)
										then
											draw_sx = nil
											draw_sy = nil
										end
									else
										draw_sx = nil
										draw_sy = nil
									end
								end
							end
						end
					end

					if draw_sx and draw_sy and draw_ex and draw_ey then
						-- translate to topleft, scale down for our rediculous draw function
						draw_sx =	((draw_sx - minX) * scale_x) / Minimap:GetWidth()
						draw_sy = (((draw_sy - minY) * scale_y) - Minimap:GetHeight()) / -Minimap:GetHeight()
						draw_ex =	((draw_ex - minX) * scale_x) / Minimap:GetWidth()
						draw_ey = (((draw_ey - minY) * scale_y) - Minimap:GetHeight()) / -Minimap:GetHeight()

						MapMarker.drawLine(Minimap, MINIMAP_LINES, USED_MINIMAP, draw_sx, draw_sy, draw_ex, draw_ey, MapMarker.getColor({"Herb"}));
						--end
					end
				end

				-- store last point
				last_x = cur_x
				last_y = cur_y
				last_inside = cur_inside
			end
		end
	end
end
throttle:SetScript("OnUpdate", updateLines) 